#include "stdafx.h"
#include "FluidSim.h"
#include "resource.h"
#include "MainFrm.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


#pragma warning(suppress: 26409 26433 26440 26477)
IMPLEMENT_DYNCREATE(CMainFrame, CFrameWnd)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CMainFrame, CFrameWnd)
  ON_REGISTERED_MESSAGE(AFX_WM_DRAW2D, &CMainFrame::OnDraw2D)
    ON_WM_CREATE()
    ON_WM_DESTROY()
    ON_WM_TIMER()
    ON_COMMAND(ID_VIEW_SIMULATION_PROPERTIES, &CMainFrame::OnViewSimulationProperties)
    ON_WM_RBUTTONDOWN()
    ON_WM_RBUTTONUP()
    ON_WM_LBUTTONDOWN()
    ON_WM_LBUTTONUP()
    ON_COMMAND(ID_VIEW_CLEAR, &CMainFrame::OnViewClear)
    ON_WM_MOUSEMOVE()
END_MESSAGE_MAP()

#pragma warning(suppress: 26455)
CMainFrame::CMainFrame() : m_nTimerID(0),
                           m_nFramesRenderedInLastSecond(0),
                           m_nLastSecond(0),
                           m_nFPS(0),
                           m_bCaptured(false),
                           m_bLButtownDown(false),
                           m_bRButtownDown(false),
                           m_omx(0),
                           m_omy(0),
                           m_mx(0),
                           m_my(0),
                           m_Simulator(theApp.m_nCellCount, theApp.m_fDiffusion, theApp.m_fDt, theApp.m_fViscosity)
{
  EnableD2DSupport();
}

LRESULT CMainFrame::OnDraw2D(WPARAM /*wParam*/, LPARAM lParam)
{
  //Pull out the render target
#pragma warning(suppress: 26429 26490)
  CHwndRenderTarget* pRenderTarget = reinterpret_cast<CHwndRenderTarget*>(lParam);
  ASSERT_VALID(pRenderTarget);

  GetFromUI();
  m_Simulator.vel_step();
  m_Simulator.dens_step();

  if (pRenderTarget->IsValid())
  {
    ++m_nFramesRenderedInLastSecond;

    pRenderTarget->SetDpi(CD2DSizeF(96, 96));
#pragma warning(suppress: 26472)
    pRenderTarget->Clear(D2D1::ColorF(static_cast<UINT32>(RGB(0, 0, 0))));

    //Draw the density of each cell
    const int N = m_Simulator.N();
    for (int i=1; i<=N; i++)
    {
      for (int j=1; j<=N; j++)
      {
        const int nLeft = (i - 1) * theApp.m_nCellSize;
        const int nTop = (j - 1) * theApp.m_nCellSize;
        const CD2DRectF rect(static_cast<FLOAT>(nLeft), static_cast<FLOAT>(nTop), static_cast<FLOAT>(nLeft + theApp.m_nCellSize), static_cast<FLOAT>(nTop + theApp.m_nCellSize));
#pragma warning(suppress: 26446)
        const float d = m_Simulator.m_dens[CFluidSimulator::IX(N, i, j)];
        CD2DSolidColorBrush brush(pRenderTarget, D2D1::ColorF(d, d, d));
        pRenderTarget->FillRectangle(rect, &brush);
      }
    }

    //Draw the velocity field if required
    if (theApp.m_bVelocityField)
    {
      CD2DSolidColorBrush brush(pRenderTarget, D2D1::ColorF::White);
      for (int i=1; i<=N; i++)
      {
        for (int j=1; j<=N; j++)
        {
          const CD2DPointF ptFrom((i - 0.5f) * theApp.m_nCellSize, (j - 0.5f) * theApp.m_nCellSize);
          CD2DPointF ptTo(ptFrom);
#pragma warning(suppress: 26446)
          ptTo.x += m_Simulator.m_u[CFluidSimulator::IX(N, i, j)] * N;
#pragma warning(suppress: 26446)
          ptTo.y += m_Simulator.m_v[CFluidSimulator::IX(N, i, j)] * N;
          pRenderTarget->DrawLine(ptFrom, ptTo, &brush);
        }
      }
    }

    //Draw the FPS value
    CRect rClient;
    GetClientRect(rClient);
    CString sFPS;
#pragma warning(suppress: 26472)
    sFPS.Format(_T("FPS: %d"), static_cast<int>(m_nFPS));
    CD2DSolidColorBrush brush(pRenderTarget, D2D1::ColorF::AliceBlue);
    pRenderTarget->DrawText(sFPS, CD2DRectF(rClient), &brush);

    //Update the stats for calculating the FPS
    const ULONGLONG nCurrentTickCount = GetTickCount64();
    if (m_nLastSecond == 0)
        m_nLastSecond = nCurrentTickCount;
    if ((nCurrentTickCount - m_nLastSecond) >= 1000)
    {
      m_nFPS = m_nFramesRenderedInLastSecond;
      m_nFramesRenderedInLastSecond = 0;
      m_nLastSecond = nCurrentTickCount;
    }
  }

  return TRUE;
}

#pragma warning(suppress: 26434)
int CMainFrame::OnCreate(LPCREATESTRUCT lpCreateStruct)
{
  //Let the base class do its thing
  if (__super::OnCreate(lpCreateStruct) == -1)
    return -1;

  //Create the timer
  m_nTimerID = SetTimer(1, 16, nullptr);

  return 0;
}

#pragma warning(suppress: 26434)
void CMainFrame::OnDestroy()
{
  //Destroy the timer
  KillTimer(m_nTimerID);
  m_nTimerID = 0;

  //Let the base class do its thing
  __super::OnDestroy();
}

#pragma warning(suppress: 26434)
void CMainFrame::OnTimer(UINT_PTR nIDEvent)
{
  if (nIDEvent == m_nTimerID)
    Invalidate(FALSE);
  else
    __super::OnTimer(nIDEvent);
}

void CMainFrame::OnViewSimulationProperties()
{
  CPropertiesDlg dlg;
  dlg.m_nCellCount = theApp.m_nCellCount;
  dlg.m_nCellSize = theApp.m_nCellSize;
  dlg.m_fForce = theApp.m_fForce;
  dlg.m_fSource = theApp.m_fSource;
  dlg.m_fDiffusion = theApp.m_fDiffusion;
  dlg.m_fDt = theApp.m_fDt;
  dlg.m_fViscosity = theApp.m_fViscosity;
  dlg.m_bVelocityField = theApp.m_bVelocityField;
  if (dlg.DoModal() == IDOK)
  {
    theApp.m_nCellCount = dlg.m_nCellCount;
    theApp.m_nCellSize = dlg.m_nCellSize;
    theApp.m_fForce = dlg.m_fForce;
    theApp.m_fSource = dlg.m_fSource;
    theApp.m_fDiffusion = dlg.m_fDiffusion;
    theApp.m_fDt = dlg.m_fDt;
    theApp.m_fViscosity = dlg.m_fViscosity;
    theApp.m_bVelocityField = dlg.m_bVelocityField;

    theApp.WriteProfileInt(_T("General"), _T("CellCount"), theApp.m_nCellCount);
    theApp.WriteProfileInt(_T("General"), _T("CellSize"), theApp.m_nCellSize);
    CString sValue;
    sValue.Format(_T("%f"), theApp.m_fForce);
    theApp.WriteProfileString(_T("General"), _T("Force"), sValue);
    sValue.Format(_T("%f"), theApp.m_fSource);
    theApp.WriteProfileString(_T("General"), _T("Source"), sValue);
    sValue.Format(_T("%f"), theApp.m_fDiffusion);
    theApp.WriteProfileString(_T("General"), _T("Diffusion"), sValue);
    sValue.Format(_T("%f"), theApp.m_fDt);
    theApp.WriteProfileString(_T("General"), _T("DT"), sValue);
    sValue.Format(_T("%f"), theApp.m_fViscosity);
    theApp.WriteProfileString(_T("General"), _T("Viscosity"), sValue);

    m_Simulator.SetN(theApp.m_nCellCount);
    m_Simulator.SetDiffusion(theApp.m_fDiffusion);
    m_Simulator.SetViscosity(theApp.m_fViscosity);
    m_Simulator.SetDt(theApp.m_fDt);
  }
}

#pragma warning(suppress: 26434)
void CMainFrame::OnRButtonDown(UINT /*nFlags*/, CPoint point)
{
  m_bRButtownDown = true;
  m_omx = m_mx = point.x;
  m_omy = m_my = point.y;
  SetCapture();
  m_bCaptured = true;
}

#pragma warning(suppress: 26434)
void CMainFrame::OnLButtonDown(UINT /*nFlags*/, CPoint point)
{
  m_bLButtownDown = true;
  m_omx = m_mx = point.x;
  m_omy = m_my = point.y;
  SetCapture();
  m_bCaptured = true;
}

#pragma warning(suppress: 26434)
void CMainFrame::OnLButtonUp(UINT /*nFlags*/, CPoint /*point*/) noexcept
{
  m_bLButtownDown = false;
  if (m_bCaptured)
  {
    ReleaseCapture();
    m_bCaptured = false;
  }
}

#pragma warning(suppress: 26434)
void CMainFrame::OnRButtonUp(UINT /*nFlags*/, CPoint /*point*/) noexcept
{
  m_bRButtownDown = false;
  if (m_bCaptured)
  {
    ReleaseCapture();
    m_bCaptured = false;
  }
}

#pragma warning(suppress: 26434)
void CMainFrame::OnMouseMove(UINT /*nFlags*/, CPoint point) noexcept
{
  m_mx = point.x;
  m_my = point.y;
}

void CMainFrame::OnViewClear()
{
  m_Simulator.clear();
}

void CMainFrame::GetFromUI()
{
  const int N = m_Simulator.N();
  const int size = (N + 2) * (N + 2);
  for (int i=0; i<size; i++)
  {
#pragma warning(suppress: 26446)
    m_Simulator.m_u_prev[i] = 0.0f;
#pragma warning(suppress: 26446)
    m_Simulator.m_v_prev[i] = 0.0f;
#pragma warning(suppress: 26446)
    m_Simulator.m_dens_prev[i] = 0.0f;
  }
  if (!m_bLButtownDown && !m_bRButtownDown)
    return;
  const int i = (m_mx / theApp.m_nCellSize) + 1;
  const int j = (m_my / theApp.m_nCellSize) + 1;
  if ((i < 1) || (i > N) || (j < 1) || (j > N))
    return;
  if (m_bLButtownDown)
  {
#pragma warning(suppress: 26446)
    m_Simulator.m_u_prev[CFluidSimulator::IX(N, i, j)] = theApp.m_fForce * (m_mx - m_omx);
#pragma warning(suppress: 26446)
    m_Simulator.m_v_prev[CFluidSimulator::IX(N, i, j)] = theApp.m_fForce * (m_my - m_omy);
  }
  if (m_bRButtownDown)
#pragma warning(suppress: 26446)
    m_Simulator.m_dens_prev[CFluidSimulator::IX(N, i, j)] = theApp.m_fSource;
  m_omx = m_mx;
  m_omy = m_my;
}


#pragma warning(suppress: 26455)
CPropertiesDlg::CPropertiesDlg(_In_opt_ CWnd* pParent) : CDialog(CPropertiesDlg::IDD, pParent),
                                                         m_nCellCount(64),
                                                         m_nCellSize(10),
                                                         m_fForce(50.0f),
                                                         m_fSource(10.0f),
                                                         m_fDiffusion(0.00001f),
                                                         m_fDt(0.1f),
                                                         m_fViscosity(0),
                                                         m_bVelocityField(FALSE)
{
}

void CPropertiesDlg::DoDataExchange(CDataExchange* pDX)
{
  //Let the base class do its thing
  __super::DoDataExchange(pDX);

  DDX_Text(pDX, IDC_CELL_COUNT, m_nCellCount);
  DDV_MinMaxInt(pDX, m_nCellCount, 2, 1024);
  DDX_Text(pDX, IDC_CELL_SIZE, m_nCellSize);
  DDV_MinMaxInt(pDX, m_nCellSize, 1, 100);
  DDX_Text(pDX, IDC_FORCE, m_fForce);
  DDX_Text(pDX, IDC_SOURCE, m_fSource);
  DDX_Text(pDX, IDC_DIFFUSION, m_fDiffusion);
  DDX_Text(pDX, IDC_DT, m_fDt);
  DDX_Text(pDX, IDC_VISCOSITY, m_fViscosity);
  DDX_Check(pDX, IDC_VELOCITY_FIELD, m_bVelocityField);
}

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPropertiesDlg, CDialog)
END_MESSAGE_MAP()

